package sync_ratelimiter

import (
	"context"
	"fmt"
	"github.com/RussellLuo/slidingwindow"
	"github.com/redis/go-redis/v9"
	"strconv"
	"sync"
	"time"
)

type RedisDatastore struct {
	client redis.Cmdable
	ttl    time.Duration
}

func (d *RedisDatastore) fullKey(key string, start int64) string {
	return fmt.Sprintf("%s@%d", key, start)
}

func (d *RedisDatastore) Add(key string, start, value int64) (int64, error) {
	k := d.fullKey(key, start)
	c, err := d.client.IncrBy(context.Background(), k, value).Result()
	if err != nil {
		return 0, err
	}
	_, _ = d.client.Expire(context.Background(), k, d.ttl).Result()
	return c, err
}

func (d *RedisDatastore) Get(key string, start int64) (int64, error) {
	k := d.fullKey(key, start)
	value, err := d.client.Get(context.Background(), k).Result()
	if err != nil {
		if err == redis.Nil {
			err = nil
		}
		return 0, err
	}
	return strconv.ParseInt(value, 10, 64)
}

func NewRedisDatastore(client redis.Cmdable, ttl time.Duration) *RedisDatastore {
	return &RedisDatastore{client: client, ttl: ttl}
}

type IPRateLimiter struct {
	mu      sync.Mutex
	options *IPRateLimiterOptions
	store   *RedisDatastore
	ips     map[string]*Limiter
}

type IPRateLimiterOptions struct {
	windowSize    time.Duration
	maxRequests   int64
	redisAddr     string
	redisDB       int
	redisPassword string
}

func defaultOptions() *IPRateLimiterOptions {
	return &IPRateLimiterOptions{
		windowSize:    time.Minute,
		maxRequests:   60 * 10,
		redisAddr:     "localhost:6379",
		redisPassword: "",
		redisDB:       0,
	}
}

func (x *IPRateLimiterOptions) apply(opts ...Option) {
	for _, opt := range opts {
		opt(x)
	}
}

func WithWindowSize(windowSize time.Duration) Option {
	return func(o *IPRateLimiterOptions) {
		o.windowSize = windowSize
	}
}

func WithMaxRequests(maxRequests int64) Option {
	return func(o *IPRateLimiterOptions) {
		o.maxRequests = maxRequests
	}
}

func WithRedisAddr(redisAddr string) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisAddr = redisAddr
	}
}
func WithRedisDB(redisDB int) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisDB = redisDB
	}
}

func WithRedisPassword(pwd string) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisPassword = pwd
	}
}

type Option func(o *IPRateLimiterOptions)

func NewIPRateLimiter(opts ...Option) *IPRateLimiter {
	options := defaultOptions()
	options.apply(opts...)
	return &IPRateLimiter{
		mu:      sync.Mutex{},
		options: options,
		store: NewRedisDatastore(
			redis.NewClient(&redis.Options{
				Addr:     options.redisAddr,
				Password: options.redisPassword,
				DB:       options.redisDB,
			}),
			2*options.windowSize, // twice of window-size is just enough.
		),
		ips: make(map[string]*Limiter),
	}
}

type Limiter struct {
	L    *slidingwindow.Limiter
	Stop func()
}

func (x *IPRateLimiter) Allow(ip string) bool {
	x.mu.Lock()
	defer x.mu.Unlock()
	if v, ok := x.ips[ip]; ok {
		return v.L.Allow()
	} else {
		lim, stop := slidingwindow.NewLimiter(x.options.windowSize, x.options.maxRequests, func() (slidingwindow.Window, slidingwindow.StopFunc) {
			return slidingwindow.NewSyncWindow(fmt.Sprintf("SyncRateLimiter_%s", ip), slidingwindow.NewBlockingSynchronizer(x.store, 500*time.Millisecond))
		})
		x.ips[ip] = &Limiter{
			L:    lim,
			Stop: stop,
		}
		return lim.Allow()
	}
}

func (x *IPRateLimiter) Stop() {
	x.mu.Lock()
	defer x.mu.Unlock()
	for _, l := range x.ips {
		l.Stop()
	}
}
